// Reads XML format with the following DTD:
//
// <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
// <!DOCTYPE transcript [
// 	<!ELEMENT transcript (meta,subtitles)>
// 	<!ELEMENT meta (title,author?,language?,date?)>
// 	<!ELEMENT title (#PCDATA)>
// 	<!ELEMENT language (#PCDATA)>
// 	<!ELEMENT author (#PCDATA)>
// 	<!ELEMENT date (#PCDATA)>
// 	<!ELEMENT subtitles (line)*>
// 	<!ELEMENT line (timestamp,text)>
// 	<!ELEMENT timestamp (start,end)>
// 	<!ELEMENT start (#PCDATA)>
// 	<!ELEMENT end (#PCDATA)>
// 	<!ELEMENT text (#PCDATA | speaker | sfx | i)*>
// 	<!ELEMENT speaker (#PCDATA)>
// 	<!ELEMENT sfx (#PCDATA)>
// 	<!ELEMENT i (#PCDATA)>
// ]>
//
// and calculates timestamps based on the word count;
// then everything will be spewed out in SRT format (more or the less).

package main

import (
	"fmt"
	"log"
	"os"
	"regexp"
	"strings"
	"time"
	"unicode"

	"github.com/antchfx/xmlquery"
	"github.com/karrick/golf" // flag replacement library
)

// How many words can a person read per second?
const WORDS_PER_SECOND = 250.00 / 60.0

// 200 wpm = 3.333333333333333 wps
// These values are empirical — most people read between 200 to 250 words per minute.

// This represents a line in the pseudo-XML document.
type Line struct {
	TimeStamp struct {
		Start string	`xml:"start"`
		End string		`xml:"end"`
	}					`xml:"timestamp"`
	Text string			`xml:"text"`
}

// Command line variables

var (
	webvtt	bool	// user wants vtt?
	help	bool	// user invoked help request?
	reWord		= regexp.MustCompile(`[\S]+`)							// Match non-space character sequences.
	reSpeaker	= regexp.MustCompile(`(?m)(<speaker>)([\pL| |\[|\]]+)(:?<\/speaker>)`)		// Speaker becomes Voice
	reSfx		= regexp.MustCompile(`(?m)(<sfx>)([\pL| |\[|\]]+)(<\/sfx>)`)					// Special effects becomes just a class
)

func main() {
	// overide usage.
	golf.Usage = func() {
		fmt.Fprintf(os.Stderr, "Usage: \n\t%s [flags] filename\n\n", os.Args[0])
		fmt.Fprintln(os.Stderr, "If the filename is omitted, it will read from STDIN\n\nFlags:")
		golf.PrintDefaults()
	}


	// get command line arguments.
	golf.BoolVarP(&webvtt,	'v', "vtt",	 false, "enable WebVTT? (default is no, use SRT)")
	golf.BoolVarP(&help,	'h', "help", false, "show command usage")
	golf.Parse()

	if help {
		golf.Usage()
		os.Exit(0)
	}

	var f = os.Stdin	// Either STDIN or the first non-flag argument.
	var err error		// For scoping purposes.
	inputFile := golf.Arg(0)	// first non-flag argument; if non-existing, just use STDIN.

	if len(inputFile) > 0 {
		f, err = os.Open(inputFile)
		if err != nil {
			f = os.Stdin
		}
	}
	doc, err := xmlquery.Parse(f)
	if err != nil {
		log.Panicln("could not parse XML input, error was:", err)
	}
	// Debug:
/* 	log.Printf("WORDS_PER_SECOND: %f — Minute: %f - Millisecond: %f\n",
		WORDS_PER_SECOND,
		float64(time.Minute),
		float64(time.Millisecond),
	) */

	// get something to put on header:
	//
	// <meta>
	// 	<title>Made in Second Life</title>
	// 	<author>Bernhardt Drax</author>
	// 	<language>EN</language>
	// </meta>
	if webvtt {
		fmt.Print("WEBVTT")
		if meta, err := xmlquery.Query(doc, "meta"); err != nil {
			agg := fmt.Sprintf(
				"%s %s %s", // I'll do all of this properly later
				meta.SelectElement("title").InnerText(),
				meta.SelectElement("author").InnerText(),
				meta.SelectElement("language").InnerText(),
			)
			if len(agg) > 0 {
				fmt.Print("\t", agg)
			}
		}
		fmt.Println()
		fmt.Println()
	}

	var aLine Line	// one line of the document.
	for i, line := range xmlquery.Find(doc, "//subtitles/line") {
		if txt := line.SelectElement("text"); txt != nil {
			aLine.Text = parseTextLine(txt)
		}
		timestamp := line.SelectElement("timestamp")
		if timestamp != nil {
			ts := timestamp.SelectElement("start")
			if ts != nil {
				aLine.TimeStamp.Start = ts.InnerText()
			}
			ts = timestamp.SelectElement("end")
			if ts != nil {
				aLine.TimeStamp.End = ts.InnerText()
			}
		}
		// extract extraneous whitespace.
		txtLines := strings.Split(aLine.Text, "\n")
		formattedText := ""
		tempTxtLine := ""
		for _, txtLine := range txtLines {
			tempTxtLine = strings.TrimSpace(txtLine)
			if len(tempTxtLine) == 0 || unicode.IsSpace(rune(tempTxtLine[0])) {
				continue
			}
			formattedText += tempTxtLine + "\n"
		}

		// count words.
		// NOTE: at some point, we really need to collapse both functionality
		// into a single solution, to stop wasting memory & time. (gwyneth 20231128)
		// Tested at Go Playground: https://play.golang.com/p/-OgHTkMOaMa
		words := wordCount(aLine.Text)
		readingTime := float64(words) / WORDS_PER_SECOND;

		// now let's do some math here!
		// Template: 00:00.000 but Go dislikes this format; it wants 00m00s000ms instead.
		// So we have to slice it a bit! (Note how weird slices in Go work...)
		tsGo := aLine.TimeStamp.Start[0:2] + "m" + aLine.TimeStamp.Start[3:5] + "s" + aLine.TimeStamp.Start[6:] + "ms"
		startTime, err := time.ParseDuration(tsGo)
		if err != nil {
			log.Printf("[WARN] Could not parse time %q — error was %q\n", startTime, err)
		}
		etGo := startTime.Seconds() + readingTime
		etGoMin := int(etGo / 60.0)
		etGoSec := int(etGo - float64(etGoMin * 60))
		etGoMs  := (etGo - float64(etGoMin * 60 + etGoSec)) * 1000
		// log.Printf("Total seconds: %#v mins: %#v; secs: %#v; msecs: %#v\n", etGo, etGoMin, etGoSec, etGoMs)
		endTime := fmt.Sprintf("%02d:%02d.%03.0f", etGoMin, etGoSec, etGoMs)

		// see what we have so far:
		fmt.Printf("%d\n%s --> %s\n%s\n",
			i,
			aLine.TimeStamp.Start,
			endTime,
			formattedText,
		)

		aLine.TimeStamp.End = endTime
	}
	if webvtt {
		fmt.Println("NOTE The End.")
	}
} // main ends here.

// Some accessory functions here.

// Counts all words in a string, using regular expressions.
// Shamelessly copied from the Internets: https://www.dotnetperls.com/word-count-go
func wordCount(value string) int {
	// Find all matches and return count.
	results := reWord.FindAllString(value, -1)
	return len(results)
}

// Parsing is trivial for SRT, but a bit more tricky for WebVTT.
func parseTextLine(txt *xmlquery.Node) string {
	if !webvtt {
		return strings.TrimSpace(txt.InnerText())
	}
	oneLine := txt.OutputXML(false)

	// decode stupid HTML entities!
	repl := strings.NewReplacer("&#34;", "\"", "&#39;", "'")

	oneLine = repl.Replace(oneLine)
	oneLine = reSpeaker.ReplaceAllString(oneLine, `
<v $2> `)	// extra space due to stripping out things
	oneLine = reSfx.ReplaceAllString(oneLine, `<c.sfx>$2</c>`)

	return oneLine
}
